

#include "vdo.h"

extern int verbose;


/* *************************************************************************
 * __u32 frame_buffer_format (Display *d, Window w)
 *
 *
 * ************************************************************************* */
__u32 frame_buffer_format (Display *d, Window w)
{
  __u32	fbf = V4L2_PIX_FMT_RGB565;

  XWindowAttributes wa;
  XGetWindowAttributes(d, w, &wa);

  if (wa.depth == 15) 
    fbf = V4L2_PIX_FMT_RGB555;
  else if (wa.depth == 16)
    fbf = V4L2_PIX_FMT_RGB565;
  else if (wa.depth == 24) 
    fbf = V4L2_PIX_FMT_BGR32;
  else if (wa.depth == 32) 
    fbf = V4L2_PIX_FMT_BGR32;
  else {
    printf("Unrecognized display depth. You may get an X Windows error.\n");
  }
  return fbf;
}

/* *************************************************************************
 * void quit_capture (Widget w, XtPointer client, XtPointer call)
 *
 *
 * ************************************************************************* */
void quit_capture (int vid, Widget w, XtPointer client, XtPointer call)
{
  struct v4l2_performance	perf;
  
  if (vid >= 0) {
    if (ioctl(vid, VIDIOC_G_PERF, &perf) != 0)
      printf("G_PERF returned error code %d\n",errno);
    else
      printf("Captured %d frames\n", perf.frames);
  }
  close (vid);
  exit(0);
}


/* *************************************************************************
 * int BV_init_display ()
 *
 *
 * ************************************************************************* */
char *BV_init_display (XWA *xwa, int argc, char **argv, VIMAGE *vimage) 
{
  Widget toplevel;
  Widget topbox;
  Widget menubox;
  Widget quit_widget;
  XWindowAttributes wa;


  toplevel = XtAppInitialize(&xwa->app, "IARI", NULL, 0, &argc, argv, NULL, NULL, 0);

  topbox = XtVaCreateManagedWidget("topbox", boxWidgetClass, toplevel,
			    XtNorientation, XtorientVertical, XtNborderWidth,
			    0, XtNhSpace, 0, XtNvSpace, 0, NULL);
  menubox = XtVaCreateManagedWidget("menubox", boxWidgetClass, topbox,
			    XtNorientation, XtorientHorizontal, XtNborderWidth,
			    0, XtNhSpace, 4, XtNvSpace, 1, NULL);
  quit_widget = XtVaCreateManagedWidget("Close!", commandWidgetClass, menubox,
			    XtNborderWidth, 0, NULL);
  XtAddCallback(quit_widget, XtNcallback, (XtCallbackProc) exit, NULL);
  xwa->viewport = 
    XtVaCreateManagedWidget ("port", portholeWidgetClass, topbox, XtNwidth,
			     vimage->width, XtNheight, vimage->height, NULL);
  XtRealizeWidget(toplevel);

  xwa->gc = XCreateGC(XtDisplay(xwa->viewport),XtWindow(xwa->viewport),0,NULL);
  if (xwa->gc == NULL) {
    printf("XCreateGC failed.\n");
    exit (1);
  }


  XGetWindowAttributes (XtDisplay(xwa->viewport), 
			DefaultRootWindow(XtDisplay(xwa->viewport)), 
			&wa);
  if (xwa->xdepth == 0) xwa->xdepth = wa.depth;
  if (verbose) printf ("Display depth is %d\n", xwa->xdepth);

  /* Set the size of display */
  //  if ((xwa->xdepth == 24) || (xwa->xdepth == 32))
  //    vimage->size = VDO_WIDTH * VDO_HEIGHT * 4;
  //  else if (xwa->xdepth == 16)
  vimage->size = VDO_WIDTH * VDO_HEIGHT * (xwa->xdepth / 8);
  vimage->data = (char *) calloc (1, vimage->size);
  if (vimage->data == NULL) {
    printf("malloc(%d) failed\n", vimage->size);
    exit (1);
  }

  vimage->ximage = XCreateImage (XtDisplay (xwa->viewport), 
				 DefaultVisual (XtDisplay (xwa->viewport), 0),
				 (xwa->xdepth==32) ? 24 : xwa->xdepth, ZPixmap,
				 0, vimage->data, vimage->width, vimage->height,
				 16, 0);


  if (vimage->ximage == NULL) {
    printf("No XImage\n");
    exit (1);
  }

  return vimage->data;
}


/* *************************************************************************
 * int BV_display_vdoframe ()
 *
 *
 * ************************************************************************* */
void BV_display_vdoframe (XWA *xwa, VIMAGE *vimage)
{
  XEvent event;

  if (XtIsRealized(xwa->viewport)) {
    XPutImage (XtDisplay(xwa->viewport), XtWindow(xwa->viewport), xwa->gc,
	       vimage->ximage, 0, 0, 0, 0, vimage->width, vimage->height); 
  }
  else 
    XtAppPeekEvent(xwa->app, &event);

  while (XtAppPending(xwa->app)) {
    XtAppNextEvent(xwa->app, &event);
    XtDispatchEvent(&event);
  }

}


/* *************************************************************************
 * int BV_vdo_display ()
 *
 *
 * ************************************************************************* */
void BV_vdo_display (XWA *xwa, VIMAGE *vimage)
{
  XPutImage (XtDisplay(xwa->viewport), XtWindow(xwa->viewport), xwa->gc,
	     vimage->ximage, 0, 0, 0, 0, vimage->width, vimage->height); 
}






