
/* *************************************************************************
 *  File : ColorMap.c
 * -------------------------------------------------------------------------
 *  All rights are reserved by Bo-kyun Na, IntellActive Research Inc.
 * -------------------------------------------------------------------------
 * Color conversions between RGB and YUV.
 * ************************************************************************* */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "vdo.h"

unsigned char *clp;

/* Data for ConvertYUVtoRGB*/
long int crv_tab[256];
long int cbu_tab[256];
long int cgu_tab[256];

long int cgv_tab[256];
long int tab_76309[256];


/* *************************************************************************
 * Function    : 
 * ------------:------------------------------------------------------------
 * Inputs      : 
 * Outputs     : 
 * Description : 
 * ************************************************************************* */
void init_dither_tab()
{
  long int crv,cbu,cgu,cgv;
  int i;   
  	
  clp = (unsigned char *)malloc(1024);
  clp += 384;
	
  for(i=-384; i<640;  i++)
	  clp[i] = (i < 0) ? 0 : ((i > 255) ? 255 : i);
  
  crv = 104597; cbu = 132201;  /* fra matrise i global.h */
  cgu = 25675;  cgv = 53279;
  
  for (i = 0; i < 256; i++) {
    crv_tab[i] = (i-128) * crv;
    cbu_tab[i] = (i-128) * cbu;
    cgu_tab[i] = (i-128) * cgu;
    cgv_tab[i] = (i-128) * cgv;
    tab_76309[i] = 76309*(i-16);
  }
}


/* *************************************************************************
 * Function    : RGB2YUV ()
 * Description : Convert pixel format from RGB16 (RGB555) to YUV411.
 * Input       : input - pt. of input buffer
 *               y,u,v - pt.s of out buffers for lum, cr, and cb.
 *               width, height - image sizes 
 * Return      : nothing
 * Date        :
 * ************************************************************************* */
void RGB16toYUV(unsigned char *input, unsigned char *y, unsigned char *u, 
		unsigned char *v, int width, int height)
{
  register int i, j;
  register unsigned char *in;
  register short cru, crv, lum;
  register unsigned char *y1 = y, *u1 = u, *v1 = v; 
  const int QCIFwidth = 176;
  const int QCIFheigth = 144;
  static int first = 1;
  static float  mult299[1024], mult587[1024], mult114[1024];
  static float  mult16874[1024], mult33126[1024], mult5[1024];
  static float mult41869[1024], mult08131[1024];
  register unsigned char r, g, b;

  in = input + width * (height - 1) * 2;

  if (first) {
    register int index;
    register int maxValue;

    maxValue = 256;

    for (index = 0; index <= maxValue; index++) {
      mult299[index] = index * 0.29900;
      mult587[index] = index * 0.58700;
      mult114[index] = index * 0.11400;
      mult16874[index] = -0.16874 * index;
      mult33126[index] = -0.33126 * index;
      mult5[index] = index * 0.50000;
      mult41869[index] = -0.41869 * index;
      mult08131[index] = -0.08131 * index;
    }
 
    first = 0;
  }

  for(i = 0; i < QCIFheigth; i++) {
    for(j = 0; j < QCIFwidth; j++) {
      b = (*(in) & 31) << 3;
      g = ((*(in) & 224) >> 2) | (*(in + 1) & 3) << 6;
      r = (*(in + 1) & 124) << 1;

      lum =  mult114[b] + mult587[g] + mult299[r];
   
      *(y1++) = (219 * lum / 255) + 16;

      if(((i % 2) == 1) && ((j % 2) == 1)) {
	cru =  mult5[b] + mult33126[g] + mult16874[r];

	*(u1++) = (224 * cru / 255) + 128;

	crv =  mult08131[b] + mult41869[g] + mult5[r];

	*(v1++) = (224 * crv / 255) + 128;
      }

      in += 2;
    }
    in -= 2 * (width + QCIFwidth);
  }
}


/* *************************************************************************
 * Function    : RGB16toRGB32 ()
 * Description : Convert pixel format from RGB16 (RGB555) to RGB32.
 * Input       : src - pt. of input buffer
 *               dst - pt. of output buf
 *               width, height - image sizes 
 * Return      : nothing
 * Date        : 1999.12.16
 * ************************************************************************* */
void RGB16toRGB32(unsigned char *src, unsigned char *dst, int width, int height)
{
  int i;
  register unsigned char *in = src;

  for (i = 0; i < width * height; i ++) {
    *(dst++) = (*(in) & 31) << 3;                           /* Blue */
    *(dst++) = ((*(in) & 224) >> 2) | (*(in + 1) & 3) << 6; /* Green */
    *(dst++) = (*(in + 1) & 124) << 1;                      /* Red */
    *(dst++) = 0;		                            /* alpha */

    in += 2;
  }
}


/* *************************************************************************
 * Function    : RGB32toRGB16 ()
 * Description : Convert pixel format to RGB16 (RGB565) from RGB32.
 * Input       : src - pt. of input buffer
 *               dst - pt. of output buf
 *               width, height - image sizes 
 * Return      : nothing
 * Date        : 1999.12.16
 * ************************************************************************* */
void RGB32toRGB16(unsigned char *src, unsigned char *dst, int width, int height)
{
  int i;
  register unsigned char *in = src;

  for (i = 0; i < width * height; i ++) {
    /* RGB32 (BGRa) into RGB555 : (G2G1G0B4B3B2B1B0aR4R3R2R1R0G4G3) */
    //    *(dst++) = (*in >> 3) | ((*(++in) & 0x38) << 2);
    //    *(dst++) = ((*in & 0xc0) >> 6) | ((*(++in) & 0xf8) >> 1);

    /* RGB32 (BGRa) into RGB565 : (G2G1G0B4B3B2B1B0R4R3R2R1R0G5G4G3) */
    *(dst++) = (*in >> 3) | ((*(++in) & 0x38) << 3);
    *(dst++) = ((*in & 0xe0) >> 5) | (*(++in) & 0xf8);

    in +=2;			/* to remove alpha band */
  }
}


/* *************************************************************************
 * Function    : RGB555to565 ()
 * Description : Convert pixel format to RGB16 (RGB565) from RGB16 (555).
 * Input       : src - pt. of input buffer
 *               dst - pt. of output buf
 *               width, height - image sizes 
 * Return      : nothing
 * Date        : 1999.12.16
 * ************************************************************************* */
void RGB555to565(unsigned char *src, unsigned char *dst, int width, int height)
{
  int i;
  register unsigned char *in = src;

  for (i = 0; i < width * height; i ++) {
    /* from RGB555 : (G2G1G0B4B3B2B1B0aR4R3R2R1R0G4G3) */
    /* into RGB565 : (G2G1G0B4B3B2B1B0R4R3R2R1R0G5G4G3) */

    *(dst++) = * in;
    *(dst++) = (*(++in) << 1) | (*in << 1);

    in ++;
  }
}
   
         
/**********************************************************************
 *
 *	Name:	         YUV411toRGB32	
 *	Description:     Converts YUV image to RGB 32
 *	
 *	Input:	         pointer to source luma, Cr, Cb, destination,
 *                       image width and height
 *	Returns:       
 *	Side effects:
 *
 *	Date: 951208	Author: Karl.Lillevold@nta.no
 *					Modified 991104 Tom Stachowiak
 *
 ***********************************************************************/
void YUV411toRGB32 (unsigned char *src0,unsigned char *src1,unsigned char *src2,
		   unsigned char *dst_ori,int width,int height)
{       
  extern long int crv_tab[];
  extern long int cbu_tab[];
  extern long int cgu_tab[];

  extern long int cgv_tab[];
  extern long int tab_76309[];

  int y11,y21;
  int y12,y22;
  int y13,y23;
  int y14,y24;
  int u,v; 
  int i,j;
  int c11, c21, c31, c41;
  int c12, c22, c32, c42;
  //  unsigned int DW;
  unsigned int *id1, *id2;
  unsigned char *py1,*py2,*pu,*pv;
  unsigned char *d1, *d2;
  
  d1 = dst_ori;
  d1 += width*height*4 - width*4;
  d2 = d1 - width*4;
  
  py1 = src0; pu = src1; pv = src2;
  py2 = py1 + width;
 
  id1 = (unsigned int *)d1;
  id2 = (unsigned int *)d2;

  for (j = 0; j < height; j += 2) { 
    /* line j + 0 */
    for (i = 0; i < width; i += 4) {
      u = *pu++;
      v = *pv++;
      c11 = crv_tab[v];
      c21 = cgu_tab[u];
      c31 = cgv_tab[v];
      c41 = cbu_tab[u];
      u = *pu++;
      v = *pv++;
      c12 = crv_tab[v];
      c22 = cgu_tab[u];
      c32 = cgv_tab[v];
      c42 = cbu_tab[u];

      y11 = tab_76309[*py1++]; /* (255/219)*65536 */
      y12 = tab_76309[*py1++];
      y13 = tab_76309[*py1++]; /* (255/219)*65536 */
      y14 = tab_76309[*py1++];

      y21 = tab_76309[*py2++];
      y22 = tab_76309[*py2++];
      y23 = tab_76309[*py2++];
      y24 = tab_76309[*py2++];

      *d1++ = clp[(y11 + c41)>>16]; /* B */
      *d1++ = clp[(y11 - c21 - c31)>>16]; /* G */
      *d1++ = clp[(y11 + c11)>>16]; /* R */
      *d1++ = 0;

      *d1++ = clp[(y12 + c41)>>16]; /* B */
      *d1++ = clp[(y12 - c21 - c31)>>16]; /* G */
      *d1++ = clp[(y12 + c11)>>16]; /* R */
      *d1++ = 0;

      *d1++ = clp[(y13 + c42)>>16]; /* B */
      *d1++ = clp[(y13 - c22 - c32)>>16]; /* G */
      *d1++ = clp[(y13 + c12)>>16]; /* R */
      *d1++ = 0;

      *d1++ = clp[(y14 + c42)>>16]; /* B */
      *d1++ = clp[(y14 - c22 - c32)>>16]; /* G */
      *d1++ = clp[(y14 + c12)>>16]; /* R */
      *d1++ = 0;

      *d2++ = clp[(y21 + c41)>>16]; /* B */
      *d2++ = clp[(y21 - c21 - c31)>>16]; /* G */
      *d2++ = clp[(y21 + c11)>>16]; /* R */
      *d2++ = 0;

      *d2++ = clp[(y22 + c41)>>16]; /* B */
      *d2++ = clp[(y22 - c21 - c31)>>16]; /* G */
      *d2++ = clp[(y22 + c11)>>16]; /* R */
      *d2++ = 0;

      *d2++ = clp[(y23 + c42)>>16]; /* B */
      *d2++ = clp[(y23 - c22 - c32)>>16]; /* G */
      *d2++ = clp[(y23 + c12)>>16]; /* R */
      *d2++ = 0;

      *d2++ = clp[(y24 + c42)>>16]; /* B */
      *d2++ = clp[(y24 - c22 - c32)>>16]; /* G */
      *d2++ = clp[(y24 + c12)>>16]; /* R */
      *d2++ = 0;
    }

    d1 -= (12 * width);
    d2 -= (12 * width);
    py1 += width;
    py2 += width;
  }           
}  


/**********************************************************************
 * Name        :  YUV411toRGB24 ()
 * Description :  Converts YUV411 image to RGB24 (packed mode)
 * Input       :  Pointer to source luma, Cr, Cb, destination,
 *                image width and height
 * Returns     :       
 * Side effects:
 * Date        :  951208	Author: Karl.Lillevold@nta.no
 ***********************************************************************/
void YUV411toRGB24 (unsigned char *src0, unsigned char *src1, 
 	    unsigned char *src2, unsigned char *dst_ori, int width, int height)
{       
  extern long int crv_tab[];
  extern long int cbu_tab[];
  extern long int cgu_tab[];
  extern long int cgv_tab[];
  extern long int tab_76309[];

  int y11,y21;
  int y12,y22;
  int y13,y23;
  int y14,y24;
  int u,v; 
  int i,j;
  int c11, c21, c31, c41;
  int c12, c22, c32, c42;
  unsigned int DW;
  unsigned int *id1, *id2;
  unsigned char *py1,*py2,*pu,*pv;
  unsigned char *d1, *d2;
  
  d1 = dst_ori;
  d1 += width*height*3 - width*3;
  d2 = d1 - width*3;
  
  py1 = src0; pu = src1; pv = src2;
  py2 = py1 + width;
 
  id1 = (unsigned int *)d1;
  id2 = (unsigned int *)d2;

  for (j = 0; j < height; j += 2) { 
    /* line j + 0 */
    for (i = 0; i < width; i += 4) {
      u = *pu++;
      v = *pv++;
      c11 = crv_tab[v];
      c21 = cgu_tab[u];
      c31 = cgv_tab[v];
      c41 = cbu_tab[u];
      u = *pu++;
      v = *pv++;
      c12 = crv_tab[v];
      c22 = cgu_tab[u];
      c32 = cgv_tab[v];
      c42 = cbu_tab[u];

      y11 = tab_76309[*py1++]; /* (255/219)*65536 */
      y12 = tab_76309[*py1++];
      y13 = tab_76309[*py1++]; /* (255/219)*65536 */
      y14 = tab_76309[*py1++];

      y21 = tab_76309[*py2++];
      y22 = tab_76309[*py2++];
      y23 = tab_76309[*py2++];
      y24 = tab_76309[*py2++];

      /* RGBR*/
      DW = ((clp[(y11 + c41)>>16])) |
           ((clp[(y11 - c21 - c31)>>16])<<8) |
	   ((clp[(y11 + c11)>>16])<<16) |  
	   ((clp[(y12 + c41)>>16])<<24);
      *id1++ = DW;

      /* GBRG*/
      DW = ((clp[(y12 - c21 - c31)>>16])) |
           ((clp[(y12 + c11)>>16])<<8) |  
           ((clp[(y13 + c42)>>16])<<16) |
           ((clp[(y13 - c22 - c32)>>16])<<24);
      *id1++ = DW;

      /* BRGB*/
      DW = ((clp[(y13 + c12)>>16])) |  
           ((clp[(y14 + c42)>>16])<<8) |
           ((clp[(y14 - c22 - c32)>>16])<<16) |
           ((clp[(y14 + c12)>>16])<<24);  
      *id1++ = DW;

      /* RGBR*/
      DW = ((clp[(y21 + c41)>>16])) |
           ((clp[(y21 - c21 - c31)>>16])<<8) |
           ((clp[(y21 + c11)>>16])<<16) |  
           ((clp[(y22 + c41)>>16])<<24);
      *id2++ = DW;

      /* GBRG*/
      DW = ((clp[(y22 - c21 - c31)>>16])) |
           ((clp[(y22 + c11)>>16])<<8) |  
           ((clp[(y23 + c42)>>16])<<16) |
           ((clp[(y23 - c22 - c32)>>16])<<24);
      *id2++ = DW;

      /* BRGB*/
      DW = ((clp[(y23 + c12)>>16])) |  
           ((clp[(y24 + c42)>>16])<<8) |
           ((clp[(y24 - c22 - c32)>>16])<<16) |
           ((clp[(y24 + c12)>>16])<<24);  
      *id2++ = DW;
    }
    id1 -= (9 * width)>>2;
    id2 -= (9 * width)>>2;
    py1 += width;
    py2 += width;
  }           
}  


/**********************************************************************
 * Name        :  YUV411toRGB16 ()
 * Description :  Converts YUV411 image to RGB16 (packed mode)
 * Input       :  Pointer to source luma, Cr, Cb, destination,
 *                image width and height
 * Returns     :       
 * Side effects:
 * Date        :  951208	Author: Karl.Lillevold@nta.no
 ***********************************************************************/
void YUV411toRGB16 (unsigned char *src0,unsigned char *src1,unsigned char *src2,
		    unsigned char *dst, int width, int height)
{       
  extern long int crv_tab[];
  extern long int cbu_tab[];
  extern long int cgu_tab[];
  extern long int cgv_tab[];
  extern long int tab_76309[];

  int y11,y21;
  int y12,y22;
  int y13,y23;
  int y14,y24;
  int u,v; 
  int i,j;
  int c11, c21, c31, c41;
  int c12, c22, c32, c42;
  unsigned int DW;
  unsigned int *id1, *id2;
  unsigned char *py1,*py2,*pu,*pv;
  unsigned char *d1, *d2;
  
  d1 = dst;
  d1 += width*height*3 - width*3;
  d2 = d1 - width*3;
  
  py1 = src0; pu = src1; pv = src2;
  py2 = py1 + width;
 
  id1 = (unsigned int *)d1;
  id2 = (unsigned int *)d2;

  for (j = 0; j < height; j += 2) { 
    /* line j + 0 */
    for (i = 0; i < width; i += 4) {
      u = *pu++;
      v = *pv++;
      c11 = crv_tab[v];
      c21 = cgu_tab[u];
      c31 = cgv_tab[v];
      c41 = cbu_tab[u];
      u = *pu++;
      v = *pv++;
      c12 = crv_tab[v];
      c22 = cgu_tab[u];
      c32 = cgv_tab[v];
      c42 = cbu_tab[u];

      y11 = tab_76309[*py1++]; /* (255/219)*65536 */
      y12 = tab_76309[*py1++];
      y13 = tab_76309[*py1++]; /* (255/219)*65536 */
      y14 = tab_76309[*py1++];

      y21 = tab_76309[*py2++];
      y22 = tab_76309[*py2++];
      y23 = tab_76309[*py2++];
      y24 = tab_76309[*py2++];

      /* RGBR*/
      DW = ((clp[(y11 + c41)>>16])) |
           ((clp[(y11 - c21 - c31)>>16])<<5) |
	   ((clp[(y11 + c11)>>16])<<10) |  
	   ((clp[(y12 + c41)>>16])<<15);
      *id1++ = DW;

      /* GBRG*/
      DW = ((clp[(y12 - c21 - c31)>>16])) |
           ((clp[(y12 + c11)>>16])<<5) |  
           ((clp[(y13 + c42)>>16])<<10) |
           ((clp[(y13 - c22 - c32)>>16])<<15);
      *id1++ = DW;

      /* BRGB*/
      DW = ((clp[(y13 + c12)>>16])) |  
           ((clp[(y14 + c42)>>16])<<5) |
           ((clp[(y14 - c22 - c32)>>16])<<10) |
           ((clp[(y14 + c12)>>16])<<15);  
      *id1++ = DW;

      /* RGBR*/
      DW = ((clp[(y21 + c41)>>16])) |
           ((clp[(y21 - c21 - c31)>>16])<<5) |
           ((clp[(y21 + c11)>>16])<<10) |  
           ((clp[(y22 + c41)>>16])<<15);
      *id2++ = DW;

      /* GBRG*/
      DW = ((clp[(y22 - c21 - c31)>>16])) |
           ((clp[(y22 + c11)>>16])<<5) |  
           ((clp[(y23 + c42)>>16])<<10) |
           ((clp[(y23 - c22 - c32)>>16])<<15);
      *id2++ = DW;

      /* BRGB*/
      DW = ((clp[(y23 + c12)>>16])) |  
           ((clp[(y24 + c42)>>16])<<5) |
           ((clp[(y24 - c22 - c32)>>16])<<10) |
           ((clp[(y24 + c12)>>16])<<15);  
      *id2++ = DW;
    }
    id1 -= (9 * width)>>2;
    id2 -= (9 * width)>>2;
    py1 += width;
    py2 += width;
  }           
}


/* *************************************************************************
 * Name        : YUV411toRGB24sol ()
 * -------------------------------------------------------------------------
 * Input       : image *src, image *dst, int depth
 * Output      :
 * Description : Convert a 4:2:0 YCbCr 601 (or YUV411) image into an RGB 
 *               gamma image (nonlinear RGB, CCIR709), which is of xx bit
 *               pixels and laid out in an BGR format.
 * SideEffect  :
 * Author      : Ihtishm Kabir
 * Modified by : Bo-kyun Na (bkna@ecs.syr.edu), IntellActive Research Lab.
 * Date        :
 * ************************************************************************* */

//#define CLIP(x, y) if(x<0) y=0; else if(x>255) y=255; else y=(unsigned char)x;
#define CLIP(x, y) (y = (x < 0) ? 0 : (x > 255) ? 255 : (unsigned char) x)

/*
 * The width and height of U and V are guaranteed to be half 
 * of  those of Y and dst image
 */

/* Conversion Equation:
 *   R = 1.169 * Y              + 1.6    * Cr - 223.5
 *   G = 1.169 * Y + 2.03  * Cb               - 278.77
 *   B = 1.169 * Y - 0.394 * Cb - 0.8144 * Cr + 136
 */

#define K00 1.169
#define K02 1.6
#define K03 -223.5
#define K11 -.394
#define K12 -.8144
#define K13 136.
#define K21 2.03
#define K23 -278.77

/* type define of image */
// typedef struct {
//  int fd		/* called as file descripter or file identifier */
//  unsigned char *data;
//  unsigned int width;
//  unsigned int height;
//  unsigned int depth;	/* # of bits used to digitally represent the pixel. */
//  unsigned int plane;	/* 1-monochrome (grayscaled), 3-color, and so on. */
//  unsigned int pixel_stride;	/* The orther to store:                  */ 
//				/* R0G0B0R1G1B1.. or R0R1..G0..B0B1B2... */
//  unsigned int linebyte; /* # of bytes betw. the same col. positions */
//			   /* in successive rows.                      */
// } image;


void YUV411toRGB24sol(unsigned char *src0,unsigned char *src1,unsigned char *src2,
		    unsigned char *dst, int width, int height)
{
  register unsigned char *y, *u, *v, *d;
  register float y1, u1, u2, v1, v2;
  register float result;
  unsigned char *y_sav, *u_sav, *v_sav, *d_sav;
  unsigned int rows, cols;
  unsigned int ylb, ulb, vlb, dlb;

  int i, j;


  /* get parameters from the images */  
  y_sav = y = src0;
  u_sav = u = src1;
  v_sav = v = src2; 
  d_sav = d = dst;

  dlb = width;
  ylb = width;
  ulb = width;
  vlb = width;

  rows = height / 2;
  cols = width / 2;

  for (i = 0; i < rows; i ++) {
    for (j = 0; j < cols; j++) {
printf ("OK:0: rows %d, col %d\n", i, j);      
      /* compute the partial u, v. */
      u1 = K11 * (float) (*u);
      u2 = K21 * (float) (*u) + K23; 
      v1 = K02 * (float) (*v) + K03;
      v2 = K12 * (float) (*v) + K13;
printf ("OK:1\n");      
      /* *** computation four output pixels at a time *** */

      /* top left */
      y1 = K00 * (float) * y;
      result = y1 + v1;
      CLIP (result, *(d + 2));
      result = y1 + u1 + v2;
      CLIP (result, *(d + 1));
      result = y1 + u2;
      CLIP (result, *d);

      /* top right */
      y1 = K00 * (float) *(y + 1);
      result = y1 + v1;
      CLIP (result, *(d + 5));
      result = y1 + u1 + v2;
      CLIP (result, *(d + 4));
      result = y1 + u2;
      CLIP (result, *(d + 3));

      /* bottom left */
      y1 = K00 * (float) *(y + ylb);
      result = y1 + v1;
      CLIP (result, *(d + dlb + 2));
      result = y1 + u1 + v2;
      CLIP (result, *(d + dlb + 1));
      result = y1 + u2;
      CLIP (result, *(d + dlb));

      /* bottom right */
      y1 = K00 * (float) *(y + ylb + 1);
      result = y1 + v1;
      CLIP (result, *(d + dlb + 5));
      result = y1 + u1 + v2;
      CLIP (result, *(d + dlb + 4));
      result = y1 + u2;
      CLIP (result, *(d + dlb + 3));

      /* move to next u, v pixels.... */
      u++;
      v++;
      /* the next 2x2 and dst pixel block */
      d += 6;
      y += 2;
    }
    printf ("OK:4\n");      
    /* updata all the pixel pointers */
    u = (u_sav += ulb);
    v = (v_sav += vlb);
    d = (d_sav += (dlb << 1));
    y = (y_sav += (ylb << 1));
  }
}


